#ifndef PLAYER_RECORDER_H
#define PLAYER_RECORDER_H

#include "vs10xx_uc.h"

#define SDI_MAX_TRANSFER_SIZE 32

#define min(a,b) (((a)<(b))?(a):(b))


extern int player_init(void);
extern int player_connected(void);
extern u_int16 ReadSci(u_int8 addr);
extern void WriteSci(u_int8 addr, u_int16 data);
extern int WriteSdi(const u_int8 *data, u_int16 bytes);
extern u_int16 ReadVS10xxMem(u_int16 addr);

#endif
